# config.py | Flask application configuration
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os

import sentry_sdk
from flask import Flask, request
from sentry_sdk.integrations.flask import FlaskIntegration

from sciolyid.data import logger  # pylint: disable=unused-import
import sciolyid.config as config

if config.options["sentry"]:
    sentry_sdk.init(
        release=f"{os.getenv('CURRENT_PLATFORM')} Release "
        + (
            f"{os.getenv('GIT_REV', '')[:8]}"
            if os.getenv("CURRENT_PLATFORM") != "Heroku"
            else f"{os.getenv('HEROKU_RELEASE_VERSION')}:{os.getenv('HEROKU_SLUG_DESCRIPTION')}"
        ),
        dsn=os.getenv(config.options["sentry_web_dsn_env"]),
        integrations=[FlaskIntegration()],
    )

app = Flask(__name__)
app.config["SEND_FILE_MAX_AGE_DEFAULT"] = 0
app.config["SESSION_COOKIE_SAMESITE"] = "Strict"
app.config["SESSION_COOKIE_SECURE"] = True
app.config["SESSION_COOKIE_NAME"] = config.options["name"]
app.secret_key = os.getenv(config.options["secret_key_env"])
FRONTEND_URL: str = os.getenv(config.options["frontend_url_env"], "")


@app.after_request  # enable CORS
def after_request(response):
    header = response.headers
    header["Access-Control-Allow-Origin"] = FRONTEND_URL
    header["Access-Control-Allow-Credentials"] = "true"
    header["Access-Control-Allow-Headers"] = request.access_control_request_headers
    header["Access-Control-Allow-Methods"] = "HEAD, GET, OPTIONS, POST, DELETE"
    return response



# config.py | config data
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Optional, Any

required: Dict[str, Optional[str]] = {
    "bot_description": None,  # short bot description
    "bot_signature": None,  # signature for embeds
    "prefixes": None,  # bot prefixes, primary prefix is first in list
    "id_type": None,  # stars, fossils, muscles, etc. - plural noun
    "github_image_repo_url": None,  # link to github image repo
    "support_server": None,  # link to discord support server
    "source_link": None,  # link to source code (may be hosted on github)
    "name": None,  # all lowercase, no spaces, doesn't really matter what this is
}

id_required: Dict[str, Optional[str]] = {
    "category_name": None,  # space thing, bird order, muscle group - what you are splitting groups by
}

web_required: Dict[str, Optional[str]] = {
    "client_id": None,  # discord client id
    "base_image_url": None,  # root of where images are hosted
    "validation_repo_url": None,  # github repo where images are temporarily held
}

optional: Dict[str, Any] = {
    "members_intent": False,  # whether the privileged members intent is enabled in the developer portal
    "download_func": None,  # asyncronous function that downloads images locally to download_dir
    "download_dir": "github_download/",  # local directory containing media (images)
    "data_dir": "data/",  # local directory containing the id data
    "list_dir": "lists/",  # directory within data_dir containing id lists
    "restricted_list_dir": None,  # directory within data_dir containg id lists that are available by selection only
    "wikipedia_file": "wikipedia.txt",  # filename within data_dir containing wiki urls for every item
    "meme_file": None,
    "logs": True,  # enable logging
    "log_dir": "logs/",  # directory for text logs/backups
    "bot_files_dir": "",  # folder for bot generated files (downloaded images, logs)
    "short_id_type": "",  # short (usually 1 letter) form of id_type, used as alias for the pic command
    "invite": "This bot is currently not available outside the support server.",  # bot server invite link
    "authors": "person_v1.32, hmmm, and EraserBird",  # creator names
    "id_groups": True,  # true/false - if you want to be able to select certain groups of items to id
    "category_aliases": {},  # aliases for categories
    "disable_extensions": [],  # bot extensions to disable (media, check, skip, hint, score, sessions, race, other)
    "custom_extensions": [],  # custom bot extensions to enable
    "sentry": False,  # enable sentry.io error tracking
    "local_redis": True,  # use a local redis server instead of a remote url
    "bot_token_env": "token",  # name of environment variable containing the discord bot token
    "sentry_dsn_env": "SENTRY_DISCORD_DSN",  # name of environment variable containing the sentry dsn
    "redis_env": "REDIS_URL",  # name of environment variable containing the redis database url
    "backups_channel": None,  # discord channel id to upload database backups (None/False to disable)
    "backups_dir": "backups/",  # directory to put database backup files before uploading
    "holidays": True,  # enable special features on select holidays
    "sendas": True,  # enable the "sendas" command
}

web_optional: Dict[str, Any] = {
    "tmp_upload_dir": "uploaded/",  # directory for temporary file storage
    "validation_local_dir": "validation_repo/",  # directory for cloning the validation repo
    "git_token_env": "GIT_TOKEN",  # environment variable with github auth token
    "git_user_env": "GIT_USERNAME",  # environment variable with github auth token
    "git_email_env": "GIT_EMAIL",  # environment variable with github auth token
    "validation_repo_dir": "",  # directory in validation repo to store files
    "hashes_url": [],  # urls to raw hashes.csv file in both image repos
    "ids_url": [],  # urls to raw ids.csv file in both image repos
    "commit_url_format": [],  # a format string for commit urls to both repos - image repo is first, validation repo is second
    "sentry_web_dsn_env": "SENTRY_API_DSN",  # name of environment variable containing the sentry dsn
    "celery_broker_env": "CELERY_BROKER_URL",  # name of environment variable with the database url for celery (broker)
    "secret_key_env": "FLASK_SECRET_KEY",  # name of environment variable for signed cookies secret key
    "frontend_url_env": "FRONTEND_URL",  # name of environment variable for frontend url
    "client_secret_env": "DISCORD_CLIENT_SECRET",  # name of environment variable for discord client secret
    "discord_webhook_env": "DISCORD_WEBHOOK_URL",  # webhook url for discord notification log
    "verification_server": None,  # invite to special discord server for people adding images, default to support server
}

options: Dict[str, Any] = {
    k: v
    for d in (required, id_required, web_required, optional, web_optional)
    for k, v in list(d.items())
}


class BotConfigError(Exception):
    def __init__(
        self, message="An error occurred in the config process."
    ):  # pylint: disable=useless-super-delegation
        super().__init__(message)



# git_tasks.py | Celery tasks for interacting with git
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import os
import random
import shutil
import time
from typing import Callable, Union

from celery.utils.log import get_task_logger

import sciolyid.config as config
import sciolyid.web.functions.webhooks as webhooks
from sciolyid.data import get_category
from sciolyid.web.functions.images import filename_lookup
from sciolyid.web.git import image_repo, verify_repo
from sciolyid.web.tasks import celery_app, database

logger = get_task_logger(__name__)

GIT_PUSHINFO_FLAGS = (
    "ERROR",  # 1024
    "UP_TO_DATE",  # 512
    "FAST_FORWARD",  # 256
    "FORCED_UPDATE",  # 128
    "DELETED",  # 64
    "REMOTE_FAILURE",  # 32
    "REMOTE_REJECTED",  # 16
    "REJECTED",  # 8
    "NO_MATCH",  # 4
    "NEW_HEAD",  # 2
    "NEW_TAG",  # 1
)
GIT_PUSH_OPCODES = (
    "CHECKING_OUT",  # 256
    "FINDING_SOURCES",  # 128
    "RESOLVING",  # 64
    "RECEIVING",  # 32
    "WRITING",  # 16
    "COMPRESSING",  # 8
    "COUNTING",  # 4
    "END",  # 2
    "BEGIN",  # 1
)


def _push_helper(repo, commit_message, progress=None):
    logger.info("pushing (helper)")
    repo.remote("origin").pull()
    index = repo.index
    added = index.add("*")
    logger.info(f"added {len(added)} files")
    commit = index.commit(commit_message)
    push_result = repo.remote("origin").push(progress=progress)
    if len(push_result) == 0:
        return (None, commit)
    set_flags = []
    for i, flag in enumerate(f"{push_result[0].flags:0>11b}"):
        if int(flag):
            set_flags.append(GIT_PUSHINFO_FLAGS[i])
    logger.info(set_flags)
    return (set_flags, commit, push_result[0])


@celery_app.task
def push(commit_message: str, user_id: Union[int, str]):
    logger.info("pushing!")
    result = _push_helper(verify_repo, commit_message, progress=gen_progress(user_id))[
        0
    ]
    if result is None:
        database.hset(
            f"sciolyid.upload.status:{user_id}",
            mapping={"status": json.dumps(["FAIL"]), "end": int(time.time())},
        )
        logger.error("push operation failed completely!")
        webhooks.send(
            "error", message="Push for verify repo on image addition failed completely!"
        )
    else:
        database.hset(
            f"sciolyid.upload.status:{user_id}",
            mapping={"status": json.dumps(result), "end": int(time.time())},
        )
    database.delete(f"sciolyid.upload.save:{user_id}")
    database.expire(f"sciolyid.upload.status:{user_id}", 60)


def gen_progress(user_id: Union[int, str]) -> Callable:
    if isinstance(user_id, int):
        user_id = str(user_id)

    def wrapped_progress(op_code, cur_count, max_count=None, message=""):
        nonlocal user_id
        readable_opcode = {
            GIT_PUSH_OPCODES[i] if int(code) else None
            for i, code in enumerate(f"{op_code:0>9b}")
        }
        readable_opcode.discard(None)
        data = {
            "op_code": json.dumps(list(readable_opcode)),
            "cur_count": json.dumps(cur_count),
            "max_count": json.dumps(max_count),
            "message": json.dumps(message),
        }
        database.hset(f"sciolyid.upload.status:{user_id}", mapping=data)
        if (
            random.randint(1, 4) == 1  # 25%
            or "BEGIN" in readable_opcode
            or "END" in readable_opcode
        ):
            # only log occasionally
            logger.info(data)

    return wrapped_progress


@celery_app.task
def move_images():
    logger.info("checking for move")

    verify_repo.remote("origin").pull()
    image_repo.remote("origin").pull()

    root = os.path.abspath(
        config.options["validation_local_dir"] + config.options["validation_repo_dir"]
    )
    lookup = filename_lookup(root)
    delete = set(
        map(
            lambda x: x.decode("utf-8"),
            database.zrangebyscore("sciolyid.verify.images:invalid", 3, "+inf")
            + database.zrangebyscore("sciolyid.verify.images:duplicate", 3, "+inf"),
        )
    )
    if delete:
        database.zremrangebyscore("sciolyid.verify.images:invalid", 3, "+inf")
        database.zremrangebyscore("sciolyid.verify.images:duplicate", 3, "+inf")
        for image in delete:
            verify_repo.index.remove(lookup[image], working_tree=True)

    valid = set(
        map(
            lambda x: x.decode("utf-8"),
            database.zrangebyscore("sciolyid.verify.images:valid", 3, "+inf"),
        )
    )
    added_items = []
    if valid:
        database.zremrangebyscore("sciolyid.verify.images:valid", 3, "+inf")
        for image in valid:
            if image in delete:
                continue
            path = lookup[image]
            item = os.path.dirname(os.path.relpath(path, root))
            added_items.append(item)
            category = get_category(item)
            shutil.copy(
                path, os.path.join(image_repo.working_tree_dir, category, item, "")
            )
            verify_repo.index.remove(path, working_tree=True)

    if valid or delete:
        verify_push = _push_helper(verify_repo, "Update through verification!")
        image_push = _push_helper(image_repo, "Update through verification!")

        ok = 0
        urls = []
        for i, result in enumerate(
            (("image repo", image_push), ("verify repo", verify_push))
        ):
            if result[1][0] is None:
                logger.info(result[0] + " failed completely!")
                webhooks.send(
                    "error",
                    message=f"Push for {result[0]} on valid image transfer failed completely!",
                )
            else:
                logger.info(result[0] + " push success!")
                ok += 1
                urls.append(
                    config.options["commit_url_format"][i].format(
                        id=result[1][1].hexsha
                    )
                )

        if ok == 2:
            webhooks.send(
                "valid",
                added=len(valid),
                rejected=len(delete),
                items=added_items,
                urls=urls,
            )
    else:
        logger.info("no changes to update!")



# stats.py | commands for bot statistics
# Copyright (C) 2019-2021  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
from io import BytesIO, StringIO

import discord
import numpy as np
import pandas as pd
from discord.ext import commands

import sciolyid.config as config
from sciolyid.data import database, logger
from sciolyid.functions import CustomCooldown, send_leaderboard, fetch_get_user


class Stats(commands.Cog):
    def __init__(self, bot):
        self.bot = bot

    @staticmethod
    def generate_series(database_key):
        """Generates a pandas.Series from a Redis sorted set."""
        logger.info("generating series")
        data = database.zrevrangebyscore(database_key, "+inf", "-inf", withscores=True)
        return pd.Series(
            {e[0]: e[1] for e in map(lambda x: (x[0].decode("utf-8"), int(x[1])), data)}
        )

    @staticmethod
    def generate_dataframe(database_keys, titles):
        """Generates a pandas.DataFrame from multiple Redis sorted sets."""
        pipe = database.pipeline()
        for key in database_keys:
            pipe.zrevrangebyscore(key, "+inf", "-inf", withscores=True)
        result = pipe.execute()
        df = pd.DataFrame()
        for i, item in enumerate(result):
            df.insert(
                len(df.columns),
                titles[i],
                pd.Series(
                    {
                        e[0]: e[1]
                        for e in map(lambda x: (x[0].decode("utf-8"), int(x[1])), item)
                    }
                ),
            )
        df = df.fillna(value=0).astype(int)
        return df

    async def convert_users(self, df):
        """Converts discord user ids in DataFrames or Series indexes to usernames."""
        current_ids = df.index
        new_index = []
        for user_id in current_ids:
            user = await fetch_get_user(int(user_id), bot=self.bot, member=False)
            if user is None:
                new_index.append("User Unavailable")
            else:
                new_index.append(f"{user.name}#{user.discriminator}")
        df.index = new_index
        return df

    # give frequency stats
    @commands.command(
        help=f"- Gives info on command/{config.options['id_type']} frequencies",
        aliases=["freq"],
    )
    @commands.check(CustomCooldown(5.0, bucket=commands.BucketType.channel))
    async def frequency(self, ctx, scope="", page=1):
        logger.info("command: frequency")

        if scope in ("command", "commands", "c"):
            database_key = "frequency.command:global"
            title = "Most Frequently Used Commands"
        elif scope in (config.options["id_type"], config.options["id_type"][0]):
            database_key = "frequency.item:global"
            title = f"Most Frequent {config.options['id_type'].title()}"
        else:
            await ctx.send(
                f"**Invalid Scope!**\n*Valid Scopes:* `commands, {config.options['id_type']}`"
            )
            return

        await send_leaderboard(ctx, title, page, database_key)

    # give bot stats
    @commands.command(
        help="- Gives statistics on different topics",
        usage="[topic]",
        aliases=["stat"],
    )
    @commands.check(CustomCooldown(5.0, bucket=commands.BucketType.channel))
    async def stats(self, ctx, topic="help"):
        logger.info("command: stats")

        if topic in ("scores", "score", "s"):
            topic = "scores"
        elif topic in ("usage", "u"):
            topic = "usage"
        elif topic in ("help", ""):
            topic = "help"
        else:
            valid_topics = ("help", "scores", "usage")
            await ctx.send(
                f"**`{topic}` is not a valid topic!**\nValid Topics: `{'`, `'.join(valid_topics)}`"
            )
            return

        embed = discord.Embed(
            title="Bot Stats", type="rich", color=discord.Color.blue(),
        )

        if topic == "help":
            embed.description = (
                "**Available statistic topics.**\n"
                + "This command is in progress and more stats may be added. "
                + "If there is a statistic you would like to see here, "
                + "please let us know in the support server."
            )
            embed.add_field(
                name="Scores",
                value=f"`{config.options['prefixes'][0]}stats [scores|score|s]`\n*Displays stats about scores.*",
            ).add_field(
                name="Usage",
                value=f"`{config.options['prefixes'][0]}stats [usage|u]`\n*Displays stats about usage.*",
            )

        elif topic == "scores":
            embed.description = "**Score Statistics**"
            scores = self.generate_series("users:global")
            scores = scores[scores > 0]
            c, d = np.histogram(scores, bins=range(0, 1100, 100), range=(0, 1000))
            c = (c / len(scores) * 100).round(1)
            embed.add_field(
                name="Totals",
                inline=False,
                value="**Sum of top 10 user scores:** `{:,}`\n".format(
                    scores.nlargest(n=10).sum()
                )
                + "**Sum of all positive user scores:** `{:,}`\n".format(scores.sum()),
            ).add_field(
                name="Computations",
                inline=False,
                value="**Mean of all positive user scores:** `{:,.2f}`\n".format(
                    scores.mean()
                )
                + "**Median of all positive user scores:** `{:,.1f}`\n".format(
                    scores.median()
                ),
            ).add_field(
                name="Distributions",
                inline=False,
                value=f"**Number of users with scores over mean:** `{len(scores[scores > scores.mean()])}`\n"
                + "**Percentage of users with scores over mean:** `{:.1%}`".format(
                    len(scores[scores > scores.mean()]) / len(scores)
                )
                + "\n**Percentage of users with scores between:**\n"
                + "".join(
                    f"\u2192 *{d[i]}-{d[i+1]-1}*: `{c[i]}%`\n"  # \u2192 is the "Rightwards Arrow"
                    for i in range(len(c))
                ),
            )

        elif topic == "usage":
            embed.description = "**Usage Statistics**"

            today = datetime.datetime.now(datetime.timezone.utc).date()
            past_month = pd.date_range(  # pylint: disable=no-member
                today - datetime.timedelta(29), today
            ).date
            keys = list(f"daily.score:{str(date)}" for date in past_month)
            keys = ["users:global"] + keys
            titles = list(
                reversed(range(1, 32))
            )  # label columns by # days ago, today is 1 day ago
            month = self.generate_dataframe(keys, titles)
            total = month.loc[:, 31]
            month = month.loc[:, 30:1]  # remove totals column
            month = month.loc[(month != 0).any(1)]  # remove users with all 0s
            week = month.loc[:, 7:1]  # generate week from month
            week = week.loc[(week != 0).any(1)]
            today = week.loc[:, 1]  # generate today from week
            today = today.loc[today != 0]

            channels_see = len(list(self.bot.get_all_channels()))
            channels_used = int(database.zcard("score:global"))

            embed.add_field(
                name="Today (Since midnight UTC)",
                inline=False,
                value="**Accounts that answered at least 1 correctly:** `{:,}`\n".format(
                    len(today)
                )
                + "**Total birds answered correctly:** `{:,}`\n".format(today.sum()),
            ).add_field(
                name="Last 7 Days",
                inline=False,
                value="**Accounts that answered at least 1 correctly:** `{:,}`\n".format(
                    len(week)
                )
                + "**Total birds answered correctly:** `{:,}`\n".format(
                    week.sum().sum()
                ),
            ).add_field(
                name="Last 30 Days",
                inline=False,
                value="**Accounts that answered at least 1 correctly:** `{:,}`\n".format(
                    len(month)
                )
                + "**Total birds answered correctly:** `{:,}`\n".format(
                    month.sum().sum()
                ),
            ).add_field(
                name="Total",
                inline=False,
                value="**Channels the bot can see:** `{:,}`\n".format(channels_see)
                + "**Channels that have used the bot at least once:** `{:,} ({:,.1%})`\n".format(
                    channels_used, channels_used / channels_see
                )
                + "*(Note: Deleted channels or channels that the bot can't see anymore are still counted).*\n"
                + "**Accounts that have used any command at least once:** `{:,}`\n".format(
                    len(total)
                )
                + "**Accounts that answered at least 1 correctly:** `{:,} ({:,.1%})`\n".format(
                    len(total[total > 0]), len(total[total > 0]) / len(total)
                )
                + "**Users the bot can see:** `{:,}`\n".format(len(self.bot.users))
                + "**Percentage of users the bot can see that have used the bot:** `{:,.1%}`\n".format(
                    len(total) / len(self.bot.users)
                )
                + "*(Note: There may be users the bot can't see anymore).*\n",
            )

        await ctx.send(embed=embed)
        return

    # export data as csv
    @commands.command(help="- Exports bot data as a csv")
    @commands.check(CustomCooldown(60.0, bucket=commands.BucketType.channel))
    async def export(self, ctx):
        logger.info("command: export")

        files = []

        async def _export_helper(database_keys, header, filename, users=False):
            if not isinstance(database_keys, str) and len(database_keys) > 1:
                data = self.generate_dataframe(
                    database_keys, header.strip().split(",")[1:]
                )
            else:
                key = (
                    database_keys
                    if isinstance(database_keys, str)
                    else database_keys[0]
                )
                data = self.generate_series(key)
            if users:
                data = await self.convert_users(data)
            with StringIO() as f:
                f.write(header)
                data.to_csv(f, header=False)
                with BytesIO(f.getvalue().encode("utf-8")) as b:
                    files.append(discord.File(b, filename))

        logger.info("exporting freq command")
        await _export_helper(
            "frequency.command:global",
            "command,amount used\n",
            "command_frequency.csv",
            users=False,
        )

        logger.info("exporting freq item")
        await _export_helper(
            "frequency.item:global",
            f"{config.options['id_type']},amounts seen\n",
            f"{config.options['id_type']}_frequency.csv",
            users=False,
        )

        logger.info("exporting streaks")
        await _export_helper(
            ["streak:global", "streak.max:global"],
            "username#discrim,current streak,max streak\n",
            "streaks.csv",
            True,
        )

        logger.info("exporting missed")
        keys = list(
            map(
                lambda x: x.decode("utf-8"),
                database.scan_iter(match="daily.incorrect:????-??-??", count=5000),
            )
        )
        keys.sort()
        titles = ",".join(map(lambda x: x.split(":")[1], keys))
        keys = ["incorrect:global"] + keys
        await _export_helper(
            keys,
            f"{config.options['id_type']},total missed,{titles}\n",
            "missed.csv",
            users=False,
        )

        logger.info("exporting scores")
        keys = list(
            map(
                lambda x: x.decode("utf-8"),
                database.scan_iter(match="daily.score:????-??-??", count=5000),
            )
        )
        keys.sort()
        titles = ",".join(map(lambda x: x.split(":")[1], keys))
        keys = ["users:global"] + keys
        await _export_helper(
            keys, f"username#discrim,total score,{titles}\n", "scores.csv", users=True
        )

        await ctx.send(files=files)


def setup(bot):
    bot.add_cog(Stats(bot))
